open VecOps

type idx =
  [ `Char of (char * Sdlvideo.color)
  | `Block of (int * int * Sdlvideo.color * int)
  | `Bitmap of (string)
  | `Colored of (idx * Sdlvideo.color)
  | `Compose of (idx list)
  | `Move of ((int * int) * idx)
  | `Realpha of (float * idx)
  | `Combine of (idx * idx * idx)
  | `Empty
  | `TileHoriz of (idx * int)
  | `TileVert of (idx * int)
  | `MirrorHoriz of idx
  | `MirrorVert of idx
  | `NoCache of idx
  | `Surface of (Sdlvideo.surface)
      
  (* dims, function *)
  | `Function of ((int * int) * ((int * int) -> ((int * int * int) * int)))
  ]

type t = { surfaces : (idx, ((int * int) * Sdlvideo.surface)) Hashtbl.t }

let char p : idx = `Char p
let block p : idx = `Block p
let file p : idx = `Bitmap p
let colored p : idx = `Colored p
let compose p : idx = `Compose p
let move p : idx = `Move p
let realpha p : idx = `Realpha p
let combine p : idx = `Combine p
let empty : idx = `Empty
let tile_horiz p : idx = `TileHoriz p
let tile_vert p : idx = `TileVert p
let mirror_horiz p : idx = `MirrorHoriz p
let mirror_vert p : idx = `MirrorVert p
let no_cache p : idx = `NoCache p

let surface p = `Surface p
let surface_ofs (ofs, idx) = `Move ofs (`Surface idx)

let of_function p : idx = `Function p

let font = lazy (Sdlttf.open_font "font.ttf" 25)

let new_surface ?like_surface (w, h) =
  match like_surface with
    | None ->
	Sdlvideo.create_RGB_surface [`SRCALPHA] ~w ~h ~bpp:32 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0xff000000l
    | Some s ->
	Sdlvideo.create_RGB_surface_format s [`SRCALPHA] ~w ~h

let get_face (ch, color) =
  let font_height = Sdlttf.font_height (Lazy.force font) in
  let s = Sdlttf.render_glyph_blended (Lazy.force font) ch color in
  let (minx, _maxx, _miny, maxy) = Sdlttf.glyph_metrics (Lazy.force font) ch in
    ((minx, font_height - maxy), s)

let create () =
  { surfaces = Hashtbl.create 100; }

let bitmaps = lazy (create ())

let get_bitmaps () = Lazy.force bitmaps

let collision_mask_of_surface s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Sdlvideo.get_pixel s x y <> 0l)
	    done )
      done );
    mask

let collision_mask_of_surface2 s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let { Sdlvideo.amask = alpha_mask } = Sdlvideo.surface_format s in
  let consider_mask = Int32.lognot alpha_mask in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Int32.logand (Sdlvideo.get_pixel s x y) consider_mask = 0l)
	    done )
      done );
    mask

let mask_with_offset : (Sdlvideo.surface -> CollisionMask.mask) -> ((int * int) * Sdlvideo.surface) -> ((int * int) * CollisionMask.mask) =
  fun f (ofs, s) ->
    (ofs, f s)

let map_surface f src =
  let w, h = Gfx.surface_dims src in
  let dst = new_surface ~like_surface:src (w, h) in
  let finish () =
    Sdlvideo.unlock src;
    Sdlvideo.unlock dst;
  in
    Sdlvideo.lock src;
    Sdlvideo.lock dst;
    try
      GobaUtils.iter2d 
	(fun (x, y) ->
	   let ((r, g, b), alpha) = Sdlvideo.get_RGBA src (Sdlvideo.get_pixel src x y) in
	     Sdlvideo.put_pixel dst x y 
	       (let (r, g, b, alpha) = 
		  f (r, g, b, alpha) in
		  Sdlvideo.map_RGB dst ~alpha (r, g, b)
	       )
	)
	((0, 0), (w, h));
      finish ();
      dst
    with exn -> 
      finish ();
      raise exn

let init_surface_pixels ?like_surface (w, h) f =
  let dst = new_surface ?like_surface (w, h) in 
  let finish () = Sdlvideo.unlock dst in
    Sdlvideo.lock dst;
    try
      GobaUtils.iter2d 
	(fun (x, y) -> Sdlvideo.put_pixel dst x y (f (x, y)))
	((0, 0), (w, h));
      finish ();
      dst
    with exn -> 
      finish ();
      raise exn

let init_surface ?like_surface (w, h) f =
  let dst = new_surface ?like_surface (w, h) in
  let finish () = Sdlvideo.unlock dst in
    Sdlvideo.lock dst;
    try
      GobaUtils.iter2d 
	(fun (x, y) -> 
	   let (color, alpha) = f (x, y) in   
	     Sdlvideo.put_pixel dst x y (Sdlvideo.map_RGB ~alpha dst color))
	((0, 0), (w, h));
      finish ();
      dst
    with exn -> 
      finish ();
      raise exn

(* oh why isn't this in the standard library.. *)
let array_for_all f a =
  let len = Array.length a in
  let rec aux idx =
    if idx >= len then
      true
    else
      if f a.(idx) then
	aux (idx + 1)
      else
	false
  in
    aux 0

let map_surfaces f srcs =
  let w, h = Gfx.surface_dims srcs.(0) in
    assert (array_for_all ((=) (w, h)) (Array.map Gfx.surface_dims srcs));
    let srcs_len = Array.length srcs in
    let dst = new_surface ~like_surface:srcs.(0) (w, h) in
    let data = Array.make srcs_len ((0, 0, 0), 0) in
      GobaUtils.iter2d 
	(fun ((x, y) as at) ->
	   for c = 0 to srcs_len - 1 do
	     data.(c) <- Sdlvideo.get_RGBA srcs.(c) (Sdlvideo.get_pixel srcs.(c) x y)
	   done;
	   Sdlvideo.put_pixel dst x y
	     (let (color, alpha) = f at data in
		Sdlvideo.map_RGB dst ~alpha color)
	)
	((0, 0), (w, h));
      dst

(* recolors white/gray *)
let recolor (dr, dg, db) ((r, g, b, a) as orig) =
  let min = List.fold_left min r [g; b] in
  let max = List.fold_left max r [g; b] in
  let delta = max - min in
    if float delta /. float max < 0.05 then
      let br = float (List.fold_left (+) r [g; b]) /. (3.0 *. 255.0) in
      let color = (int_of_float (float dr *. br), 
		   int_of_float (float dg *. br),
		   int_of_float (float db *. br),
		   a)
      in
	color
    else
      orig

let do_realpha adjust (r, g, b, a) =
  (r, g, b, min 255 (max 0 (adjust a)))

let ipolate_rgba v ((r1, g1, b1), a1) ((r2, g2, b2), a2) =
  let v' = 1.0 -. v in
  let ( *! ) r v = int_of_float (r *. float v) in
    ((v' *! r1 + v *! r2,
      v' *! g1 + v *! g2, 
      v' *! b1 + v *! b2), 
     v' *! a1 + v *! a2)

let get surface x y = Sdlvideo.get_RGBA surface (Sdlvideo.get_pixel surface x y)

(*
let dump_surface s =
  let w, h = Gfx.surface_dims s in
    for y = 0 to h - 1 do
      for x = 0 to w - 1 do
	let color, alpha = get src xc yc in
      done;
      Printf.printf "\n"
    done
*)

let blit (gfx:#Gfx.t) at (ofs, surface) =
  gfx#blit (at +| ofs) surface

(* finally had to resort to: http://en.wikipedia.org/wiki/Alpha_transparency *)
let alpha_blit ~src ~dst ~dstofs =
  let finish () = 
    Sdlvideo.unlock src;
    Sdlvideo.unlock dst
  in
  let s'w, s'h = Gfx.surface_dims src in
    Sdlvideo.lock src;
    Sdlvideo.lock dst;
    try
      for xc = 0 to s'w - 1 do
	for yc = 0 to s'h - 1 do
	  let dx, dy = xc + fst dstofs, yc + snd dstofs in
	  let src_color, src_alpha = get src xc yc in
	  let dst_color, dst_alpha = get dst dx dy in
	  let f255 v = float v /. 255.0 in
	  let new_alpha = int_of_float (min 255.0 (float src_alpha +. float dst_alpha *. (1. -. f255 src_alpha))) in 
	    if new_alpha > 1 then
	      let new_color = 
		Vector.int3 (Vector.float3 src_color *||. Vector.v3 (float src_alpha /. float new_alpha) +||.
			       Vector.float3 dst_color *||. Vector.v3 ((float dst_alpha *. (1.0 -. f255 src_alpha)) /. float new_alpha)) in
		Sdlvideo.put_pixel dst dx dy (Sdlvideo.map_RGB dst ~alpha:new_alpha new_color);
		(*	      (let r's, g's, b's = src_color in
			      let r'd, g'd, b'd = dst_color in
			      let r, g, b = new_color in
			      Printf.printf "((%d,%d,%d),%d)+((%d,%d,%d),%d)->((%d,%d,%d),%d) " 
			      r's g's b's src_alpha
			      r'd g'd b'd dst_alpha
			      r g b new_alpha)*)
	    else
	      Sdlvideo.put_pixel dst dx dy (Sdlvideo.map_RGB dst ~alpha:0 dst_color);
	done;
	(*      Printf.printf "\n";*)
      done;
      finish ()
    with exn ->
      finish ();
      raise exn


let rec bitmap ?(cache=true) (idx : idx) =
  let bm = Lazy.force bitmaps in
    match idx with
      | `Surface (surface) ->
	  ((0,0), surface)
      | `NoCache idx -> 
	  bitmap ~cache:false idx
      | `Function p ->
	  from_function p
      | other ->
	  try Hashtbl.find bm.surfaces idx
	  with Not_found ->
	    let (ofs, surface) = 
	      match other with
		| `Char el -> get_face el
		| `Block (width, height, color, alpha) ->
		    (0, 0), 
		    (let s = new_surface (width, height) in
		       Sdlvideo.fill_rect s (Sdlvideo.map_RGB ~alpha s color);
		       s)
		| `Bitmap (file) ->
		    (0, 0),
		    (Sdlloader.load_image (Printf.sprintf "gfx/%s.png" file))
		| `Colored (idx, color) ->
		    let bm = bitmap idx in
		      (fst bm, map_surface (recolor color) (snd bm))
		| `Compose (idxs) -> do_compose idxs
		| `Combine (idxs) -> do_combine idxs
		| `Move (ofs, idx) ->
		    let bm = bitmap idx in
		      (fst bm +| ofs, snd bm)
		| `Realpha (adjust, idx) ->
		    let bm = bitmap idx in
		      (fst bm, map_surface (do_realpha (fun a -> int_of_float (float a *. adjust))) (snd bm))
		| `Empty ->
		    (0, 0), Sdlvideo.create_RGB_surface [] ~w:0 ~h:0 ~bpp:24 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0x0l
		| `TileHoriz p -> do_tile_horiz p
		| `TileVert p -> do_tile_vert p
		| `MirrorHoriz p -> do_mirror_horiz p
		| `MirrorVert p -> do_mirror_vert p
		| `NoCache _ | `Surface _ | `Function _ -> failwith "argh"
	    in
	    let surface = Sdlvideo.display_format ~alpha:true surface in
	    let s = (ofs, surface) in
	      if cache then
		Hashtbl.add bm.surfaces idx s;
	      s

and do_compose idxs =
  if idxs = [] then
    bitmap `Empty
  else
    let bms = List.map bitmap idxs in
    let extreme f cmp = List.fold_left (fun v b -> if cmp v (f b) then v else (f b)) (f (List.hd bms)) (List.tl bms) in
    let min_x = extreme (fun ((x0, _), _) -> x0) (<) in
    let min_y = extreme (fun ((_, y0), _) -> y0) (<) in
    let max_x = extreme (fun ((x0, _), s) -> x0 + fst (Gfx.surface_dims s)) (>) in
    let max_y = extreme (fun ((_, y0), s) -> y0 + snd (Gfx.surface_dims s)) (>) in
    let ofs = min_x, min_y in
    let size = (max_x, max_y) -| (min_x, min_y) in
    let surface = new_surface size in
      Sdlvideo.fill_rect surface 0x00000000l;
      (*    Printf.printf "Begin composing\n";*)
      List.iter 
	(fun (ofs', bm) ->
	   let ofs = ofs' -| ofs in
	     (*	 let size = Gfx.surface_dims bm in*)
	     (*	   Sdlvideo.blit_surface 
		   ~src:bm
		   ~dst:surface
		   ~dst_rect:(Sdlvideo.rect (fst ofs) (snd ofs) (fst size) (snd size))
		   ()*)
	     alpha_blit ~src:bm ~dst:surface ~dstofs:ofs
	)
	bms;
      (*    Printf.printf "Finish composing\n%!";*)
      ofs, surface

and do_combine (value, left, right) =
  let idxs = [|value; left; right|] in
  let bms = Array.map bitmap idxs in
  let f _ vs =
    let v = 
      let ((r, g, b), _) = vs.(0) in
	float (r + g + b) /. (255.0 *. 3.0)
    in
      ipolate_rgba v vs.(1) vs.(2)
  in
    assert (array_for_all (fun (ofs, _) -> ofs = fst bms.(0)) bms);
    (fst bms.(0), 
     map_surfaces f (Array.map snd bms))

and do_tile_horiz (idx, p) = 
  let ((_, y'ofs), surface) = bitmap idx in
  let dims = Gfx.surface_dims surface in
    ((0, y'ofs), 
     init_surface_pixels ~like_surface:surface (fst dims * p, snd dims)
       (fun (x, y) -> Sdlvideo.get_pixel surface (x mod fst dims) y))

and do_tile_vert (idx, p) = 
  let ((x'ofs, _), surface) = bitmap idx in
  let dims = Gfx.surface_dims surface in
    ((x'ofs, 0), 
     init_surface_pixels ~like_surface:surface (fst dims, snd dims * p)
       (fun (x, y) -> Sdlvideo.get_pixel surface x (y mod snd dims)))

and do_mirror_horiz idx =
  let ((x'ofs, y'ofs), surface) = bitmap idx in
  let (x'size, y'size) = Gfx.surface_dims surface in
    ((-x'ofs - x'size, y'ofs),
     init_surface_pixels ~like_surface:surface (x'size, y'size)
       (fun (x, y) -> Sdlvideo.get_pixel surface (x'size - 1 - x) y))

and do_mirror_vert idx =
  let ((x'ofs, y'ofs), surface) = bitmap idx in
  let (x'size, y'size) = Gfx.surface_dims surface in
    ((x'ofs, -y'ofs - y'size),
     init_surface_pixels ~like_surface:surface (x'size, y'size)
      (fun (x, y) -> Sdlvideo.get_pixel surface x (y'size - 1 - y)))

and from_function (dims, f) =
  ((0, 0), init_surface dims f)

let dump_stats () =
  Printf.printf "Bitmaps stats: %d surfaces\n" (Hashtbl.length (Lazy.force bitmaps).surfaces)

let _ = at_exit dump_stats
